"""Contains functions to validate user input and other data."""
import re
from pathlib import Path
from typing import cast

import videoxt.constants as C
import videoxt.utils as U
from videoxt.exceptions import ValidationError


def positive_int(n: float | int | str) -> int:
    """
    Return a positive integer from a float, integer or string.

    Args:
    -----
        `n` (float | int | str): The number to validate.

    Returns:
    -----
        `int`: The number as an integer if positive.

    Raises:
    -----
        `ValidationError`: If the number is not a positive integer.
    """
    try:
        value = float(n)
    except (ValueError, TypeError):
        raise ValidationError(f"Expected integer, got {n!r}")

    if not value.is_integer():
        raise ValidationError(f"Expected integer, got {n}")

    if value <= 0:
        raise ValidationError(f"Expected positive integer, got {n}")

    return int(value)


def positive_float(n: float | int | str) -> float:
    """
    Return a positive float from a float, integer or string.

    Args:
    -----
        `n` (float | int | str): The number to validate.

    Returns:
    -----
        `float`: The number as a float if positive.

    Raises:
    -----
        `ValidationError`: If the number is not a positive float.
    """
    try:
        value = float(n)
    except (ValueError, TypeError):
        raise ValidationError(f"Expected numeric value, got {n!r}")

    if value <= 0:
        raise ValidationError(f"Expected positive number, got {n}")

    return value


def non_negative_int(n: float | int | str) -> int:
    """
    Return a non-negative integer from a float, integer or string.

    Args:
    -----
        `n` (float | int | str): The number to validate.

    Returns:
    -----
        `int`: The number as an integer if non-negative.

    Raises:
    -----
        `ValidationError`: If the number is not a non-negative integer.
    """
    try:
        value = float(n)
    except (ValueError, TypeError):
        raise ValidationError(f"Expected integer, got {n!r}")

    if not value.is_integer():
        raise ValidationError(f"Expected integer, got {n!r}")

    if value < 0:
        raise ValidationError(f"Expected non-negative integer, got {n}")

    return int(value)


def non_negative_float(n: float | int | str) -> float:
    """
    Return a non-negative float from a float, integer or string.

    Args:
    -----
        `n` (float | int | str): The number to validate.

    Returns:
    -----
        `float`: The number as a float if non-negative.

    Raises:
    -----
        `ValidationError`: If the number is not a non-negative float.
    """
    try:
        value = float(n)
    except (ValueError, TypeError):
        raise ValidationError(f"Expected numeric value, got {n!r}")

    if value < 0:
        raise ValidationError(f"Expected non-negative number, got {n}")

    return value


def valid_dir(directory: Path | str) -> Path:
    """
    Validate a path is a directory and exists and return it.

    Args:
    -----
        `directory` (pathlib.Path | str): The directory to validate.

    Returns:
    -----
        `pathlib.Path`: The directory as a Path object if valid.

    Raises:
    -----
        `ValidationError`: If the path is not a directory or does not exist.
    """
    if directory is None:
        raise ValidationError(f"Directory cannot be None, got {directory!r}")

    dir_path = Path(directory)

    if not dir_path.is_dir():
        raise ValidationError(f"Directory not found, got {directory!r}")

    if dir_path == Path("/"):
        raise ValidationError(f"Invalid directory, got {directory!r}")

    if dir_path == Path("."):
        dir_path = Path.cwd()

    if dir_path == Path(".."):
        dir_path = Path.cwd().parent

    return dir_path


def valid_filepath(filepath: Path | str, is_video: bool = False) -> Path:
    """
    Validate a path is a file and exists and return it.

    Args:
    -----
        `filepath` (pathlib.Path | str):
            The filepath to validate.
        `is_video` (bool) :
            If True, the filepath is checked against a set of supported video file
            suffixes.

    Returns:
    -----
        `pathlib.Path`: The filepath as a Path object if it is a file and exists.

    Raises:
    -----
        `ValidationError`: If the path is not a file or does not exist.
    """
    if filepath is None:
        raise ValidationError(f"Filepath cannot be None, got {filepath!r}")

    try:
        fp = Path(filepath)
    except TypeError:
        raise ValidationError(f"Invalid filepath, got {filepath!r}")

    if not fp.exists():
        raise ValidationError(f"File not found, got {filepath!r}")

    if not fp.is_file():
        raise ValidationError(f"Filepath provided is not a file, got {filepath!r}")

    if is_video:
        valid_video_file_suffix(fp.suffix)

    return fp


def valid_filename(filename: str) -> str:
    """
    Validate a filename does not contain invalid characters and return it.

    In the context of `videoxt`, a filename refers to a pathlib.Path().stem.

    Args:
    -----
        `filename` (str): The filename to validate.

    Returns:
    -----
        `str`: The filename if it does not contain invalid characters.

    Raises:
    -----
        `ValidationError`: If the filename is None, empty or contains invalid chars.
    """
    if filename is None:
        raise ValidationError(f"Invalid filename, got {filename!r}")

    if not filename:
        raise ValidationError(f"Invalid filename, got {filename!r}")

    invalid_chars = r"[\\/:*?\"<>|]"
    if re.search(invalid_chars, filename):
        raise ValidationError(
            f"Invalid filename, got {filename!r}\n"
            f"filename can't contain any of the following characters: \\/:*?\"<>|"
        )

    return filename


def valid_timestamp(timestamp: str) -> str:
    """
    Validate a timestamp is in the correct format and return it if valid.

    In the context of `videoxt` a timestamp is one typically seen during playback on
    video sharing or streaming platforms. Microseconds are truncated.

    Valid:  `M:SS`, `MM:SS`, `H:MM:SS`, `HH:MM:SS`.
    Invalid: `S`, `SS`, `H:M:S`, values greater than 59 for hours, minutes or seconds.

    Args:
    -----
        `timestamp` (str): The timestamp to validate.

    Returns:
    -----
        `str`: The timestamp if valid.

    Raises:
    -----
        `ValidationError`:
            If the timestamp is None, empty, or does not match the regex pattern.
    """
    if timestamp is None or not timestamp:
        raise ValidationError(f"Timestamp string is empty or None, got {timestamp!r}")

    timestamp = timestamp.split(".")[0]

    regex = r"^([0-9]|[0-5][0-9])(:[0-5][0-9]){1,2}$"
    if not bool(re.match(regex, timestamp)):
        raise ValidationError(
            f"Invalid timestamp format, got {timestamp!r}\n"
            f"Allowed: 'M:SS', 'MM:SS', 'H:MM:SS', 'HH:MM:SS'"
        )

    return timestamp


def valid_start_timestamp(start_timestamp: str) -> str:
    """
    Validate a start timestamp is in the correct format and return it if valid.

    Valid:  `M:SS`, `MM:SS`, `H:MM:SS`, `HH:MM:SS`.
    Invalid: `S`, `SS`, `H:M:S`, values greater than 59 for hours, minutes or seconds.

    See `valid_timestamp` for more information on how timestamps are validated.

    Args:
    -----
        `start_timestamp` (str) : The timestamp to validate (Ex: '12:34').

    Returns:
    -----
        `str`: The timestamp as a string if valid.
    """
    timestamp = valid_timestamp(start_timestamp)
    timestamp_as_seconds = U.timestamp_to_seconds(timestamp)
    timestamp_as_seconds = (
        non_negative_float(timestamp)
        if timestamp_as_seconds is None
        else non_negative_float(timestamp_as_seconds)
    )

    return timestamp


def valid_stop_timestamp(stop_timestamp: str) -> str:
    """
    Validate a stop timestamp is in the correct format and return it if valid.

    Valid:  `M:SS`, `MM:SS`, `H:MM:SS`, `HH:MM:SS`.
    Invalid: `S`, `SS`, `H:M:S`, values greater than 59 for hours, minutes or seconds.

    See `valid_timestamp` for more information on how timestamps are validated.

    Args:
    -----
        `stop_timestamp` (str): The timestamp to validate (Ex: '12:34').

    Returns:
    -----
        `str`: The timestamp as a string if valid.
    """
    timestamp = valid_timestamp(stop_timestamp)
    timestamp_as_seconds = U.timestamp_to_seconds(timestamp)
    timestamp_as_seconds = (
        positive_float(timestamp)
        if timestamp_as_seconds is None
        else positive_float(timestamp_as_seconds)
    )

    return timestamp


def valid_start_time(start_time: float | int | str) -> float | str:
    """
    Validate any start time provided is in the correct format and not negative.

    Args:
    -----
        `start_time` (float | int | str): The start time to validate.

    Returns:
    -----
        `float | str`: The start time as a float or string if valid.

    Raises:
    -----
        `ValidationError`:
            If start time is not a non-negative float or a properly formatted timestamp.
    """
    try:
        start_time_float = float(start_time)
    except ValueError:
        try:
            return valid_start_timestamp(str(start_time))
        except ValidationError:
            raise ValidationError(
                f"Invalid start time, got {start_time!r}\n"
                "Start time must be a non-negative number or a properly formatted "
                "timestamp (Ex: 'HH:MM:SS')."
            )
    else:
        try:
            return non_negative_float(start_time_float)
        except ValidationError:
            raise ValidationError(
                f"Invalid start time, got {start_time!r}\n"
                "Start time must be a non-negative number or a properly formatted "
                "timestamp (Ex: 'HH:MM:SS')."
            )


def valid_stop_time(stop_time: float | int | str) -> float | str:
    """
    Validate any stop time provided is in the correct format and not negative.

    Args:
    -----
        `stop_time` (float | int | str): The stop time to validate.

    Returns:
    -----
        `float | str`: The stop time as a float or string if valid.

    Raises:
    -----
        `ValidationError`:
            If stop time is not a positive float or a properly formatted timestamp.
    """
    try:
        stop_time_float = float(stop_time)
    except ValueError:
        try:
            return valid_stop_timestamp(str(stop_time))
        except ValidationError:
            raise ValidationError(
                f"Invalid stop time, got {stop_time!r}\n"
                "Stop time must be a positive number or a properly formatted "
                "timestamp (Ex: 'HH:MM:SS')."
            )
    else:
        try:
            return positive_float(stop_time_float)
        except ValidationError:
            raise ValidationError(
                f"Invalid stop time, got {stop_time!r}\n"
                "Stop time must be a positive number or a properly formatted "
                "timestamp (Ex: 'HH:MM:SS')."
            )


def valid_extraction_range(
    start: float, stop: float, duration: float
) -> tuple[float, float, float]:
    """
    Validate the extraction range is within bounds and return it.

    Args:
    -----
        `start` (float):
            Extraction start second.
        `stop` (float):
            Extraction stop second.
        `duration` (float):
            Duration of the video in seconds.

    Returns:
    -----
        `Tuple[float, float, float]`:
            The start, stop and duration if valid (start, stop, duration).

    Raises:
    -----
        `ValidationError`:
        - If the start second is greater than or equal to the duration.
        - If the stop second is less than or equal to the start second.
    """
    if stop > duration:
        stop = duration

    if start < 0:
        start = 0

    if start >= duration:
        raise ValidationError(
            f"Start second ({start}) is >= video duration ({duration})"
        )

    if stop <= start:
        raise ValidationError(
            f"Stop second ({stop}) must be greater than start second ({start})"
        )

    return start, stop, duration


def valid_dimensions(dimensions: tuple[int, int]) -> tuple[int, int]:
    """
    Validate values in a dimensions tuple are positive integers and return it.

    Args:
    -----
        `dimensions` (tuple[int, int]): The dimensions to validate.

    Returns:
    -----
        `tuple[int, int]`: The dimensions if both integers are positive.

    Raises:
    -----
        `ValidationError`:
            If the length of the tuple isn't 2 or one or more of the values are not
            positive integers
    """
    if len(dimensions) != 2:
        raise ValidationError(f"Invalid dimensions, got {dimensions!r}")

    dims = tuple([positive_int(dim) for dim in list(dimensions)])

    return cast(tuple[int, int], dims)


def valid_rotate_value(n: float | int | str) -> int:
    """
    Validate a rotate value is either 0, 90, 180 or 270.

    Args:
    -----
        `n` (float | int | str): The rotate value to validate.

    Returns:
    -----
        `int`: The rotate value if 0, 90, 180, or 270.

    Raises:
    -----
        `ValidationError`: If the rotate value is invalid.
    """
    try:
        val = int(n)
    except ValueError:
        raise ValidationError(
            f"Invalid rotate value, got {n!r}\n"
            f"Allowed values: {C.VALID_ROTATE_VALUES}"
        )

    if val not in C.VALID_ROTATE_VALUES:
        raise ValidationError(
            f"Invalid rotate value, got {n}\n"
            f"Allowed values: {C.VALID_ROTATE_VALUES}"
        )

    return val


def valid_audio_format(audio_format: str) -> str:
    """
    Validate audio format is supported by `videoxt` and return it if so.

    Input is converted to lowercase and stripped of any leading periods. `Mp3` and
    `.MP3` would both be considered valid and returned as `mp3`.

    See supported formats here: `videoxt.constants.SUPPORTED_AUDIO_FORMATS`.

    Args:
    -----
        `audio_format` (str): The audio format to validate.

    Returns:
    -----
        `str`: The audio format if supported.

    Raises:
    -----
        `ValidationError`: If the audio format is not supported.
    """
    fmt = audio_format.lower().lstrip(".")
    if fmt not in C.SUPPORTED_AUDIO_FORMATS:
        raise ValidationError(
            f"Unsupported audio format, got {audio_format!r}\n"
            f"Supported formats: {C.SUPPORTED_AUDIO_FORMATS}"
        )

    return fmt


def valid_image_format(image_format: str) -> str:
    """
    Validate image format is supported by `videoxt` and return it if so.

    Input is converted to lowercase and stripped of any leading periods. `jpG` and
    `.JPG` would both be considered valid and returned as `jpg`.

    See supported formats here: `videoxt.constants.SUPPORTED_IMAGE_FORMATS`.

    Args:
    -----
        `image_format` (str): The image format to validate.

    Returns:
    -----
        `str`: The image format if valid.

    Raises:
    -----
        `ValidationError`: If the image format is not supported.
    """
    fmt = image_format.lower().lstrip(".")
    if fmt not in C.SUPPORTED_IMAGE_FORMATS:
        raise ValidationError(
            f"Invalid image format, got {image_format!r}\n"
            f"Supported image formats: {C.SUPPORTED_IMAGE_FORMATS}"
        )

    return fmt


def valid_volume(volume: float | int | str) -> float:
    """
    Validate and return non-negative float audio volume. If input is negative, set to 0.

    Args:
    -----
        `volume` (float | int | str): The volume to validate.

    Returns:
    -----
        `float`: The volume as a float if valid.

    Raises:
    -----
        `ValidationError`: If the volume is None.
    """
    if volume is None:
        raise ValidationError("Volume cannot be None.")

    if isinstance(volume, (float, int)):
        return volume if volume > 0 else 0

    try:
        vol = float(volume)
    except (ValueError, TypeError):
        raise ValidationError(f"Volume expects numeric value, got {volume!r}")

    return vol if vol > 0 else 0


def valid_video_file_suffix(suffix: str) -> str:
    """
    Validate suffix provided is supported by `videoxt` and return it.

    Input is converted to lowercase and stripped of any leading periods. `Mp4` and
    `.MP4` would both be considered valid and returned as `mp4`.

    Args:
    -----
        `suffix` (str): The suffix to validate.

    Returns:
    -----
        `str`: The suffix if valid.

    Raises:
    -----
        `ValidationError`: If the video format is not supported.
    """
    sfx = suffix.lower().lstrip(".")
    if sfx not in C.SUPPORTED_VIDEO_FORMATS:
        raise ValidationError(
            f"Invalid video file suffix, got {suffix!r}\n"
            f"Supported: {C.SUPPORTED_VIDEO_FORMATS}"
        )

    return sfx


def valid_fps(fps: float | int | str) -> float:
    """
    Validate and return a positive float fps.

    Args:
    -----
        `fps` (float | int | str): The fps to validate.

    Returns:
    -----
        `float`: The fps as a float if valid.

    Raises:
    -----
        `ValidationError`: If the fps is None or not a positive float.
    """
    if fps is None:
        raise ValidationError("FPS cannot be None.")

    try:
        return positive_float(fps)
    except ValidationError:
        raise ValidationError(
            f"Invalid fps, got {fps!r}\nFPS must be a positive number."
        )


def valid_dimensions_str(dimensions: str) -> tuple[int, int]:
    # expected input is 'WxH' (Ex: '1920x1080')
    if not dimensions:
        raise ValidationError(
            f"Empty dimensions provided, got {dimensions!r}\n"
            "Expected format: 'WxH' (Ex: '1920x1080')"
        )

    if dimensions.count("x") > 1:
        raise ValidationError(
            f"Too many dimensions provided, got {dimensions!r}\n"
            "Expected format: 'WxH' (Ex: '1920x1080')"
        )

    try:
        dims = tuple(positive_int(dim) for dim in tuple(dimensions.split("x")))
    except ValidationError:
        raise ValidationError(
            f"Invalid dimensions, got {dimensions!r}\n"
            "Dimensions must be positive integers.\n"
            "Expected format: 'WxH' (Ex: '1920x1080')"
        )

    return cast(tuple[int, int], dims)


def valid_resize(resize: float | int | str) -> float:
    """
    Validate and return a positive float resize value.

    Args:
    -----
        `resize` (float | int | str): The resize value to validate.

    Returns:
    -----
        `float`: The resize value as a float if valid.

    Raises:
    -----
        `ValidationError`: If the resize value is None or not a positive float.
    """
    if resize is None:
        raise ValidationError("Resize value cannot be None.")

    try:
        return positive_float(resize)
    except ValidationError:
        raise ValidationError(
            f"Invalid resize value, got {resize!r}\n"
            "Resize value must be a positive number."
        )


def valid_speed(speed: float | int | str) -> float:
    """
    Validate and return a positive float speed value.

    Args:
    -----
        `speed` (float | int | str): The speed value to validate.

    Returns:
    -----
        `float`: The speed value as a float if valid.

    Raises:
    -----
        `ValidationError`: If the speed value is None or not a positive float.
    """
    if speed is None:
        raise ValidationError("Speed value cannot be None.")

    try:
        return positive_float(speed)
    except ValidationError:
        raise ValidationError(
            f"Invalid speed value, got {speed!r}\n"
            "Speed value must be a positive number."
        )


def valid_capture_rate(capture_rate: float | int | str) -> int:
    """
    Validate and return a positive integer capture rate.

    Args:
    -----
        `capture_rate` (float | int | str): The capture rate to validate.

    Returns:
    -----
        `int`: The capture rate as an integer if valid.

    Raises:
    -----
        `ValidationError`: If the capture rate is None or not a positive integer.
    """
    if capture_rate is None:
        raise ValidationError("Capture rate cannot be None.")

    try:
        return positive_int(capture_rate)
    except ValidationError:
        raise ValidationError(
            f"Invalid capture rate, got {capture_rate!r}\n"
            "Capture rate must be a positive integer."
        )



from pathlib import Path

import pytest

import videoxt.constants as C
from videoxt.exceptions import ValidationError
from videoxt.validators import (
    non_negative_float,
    non_negative_int,
    positive_float,
    positive_int,
    valid_audio_format,
    valid_dimensions,
    valid_dir,
    valid_extraction_range,
    valid_filename,
    valid_filepath,
    valid_image_format,
    valid_rotate_value,
    valid_start_time,
    valid_stop_time,
    valid_timestamp,
    valid_video_file_suffix,
    valid_volume,
)


def test_positive_int_valid_int():
    assert positive_int(42) == 42


def test_positive_int_valid_string():
    assert positive_int("42") == 42


def test_positive_int_valid_float():
    assert positive_int(42.0) == 42


def test_positive_int_valid_float_string():
    assert positive_int("42.0") == 42


def test_positive_int_zero_int():
    with pytest.raises(ValidationError):
        positive_int(0)


def test_positive_int_zero_int_string():
    with pytest.raises(ValidationError):
        positive_int("0")


def test_positive_int_zero_float():
    with pytest.raises(ValidationError):
        positive_int(0.0)


def test_positive_int_negative_int():
    with pytest.raises(ValidationError):
        positive_int(-42)


def test_positive_int_negative_int_string():
    with pytest.raises(ValidationError):
        positive_int("-42")


def test_positive_int_negative_float():
    with pytest.raises(ValidationError):
        positive_int(-42.0)


def test_positive_int_negative_float_string():
    with pytest.raises(ValidationError):
        positive_int("-42.0")


def test_positive_int_invalid_float():
    with pytest.raises(ValidationError):
        positive_int(42.5)


def test_positive_int_invalid_float_string():
    with pytest.raises(ValidationError):
        positive_int("42.5")


def test_positive_int_invalid_string():
    with pytest.raises(ValidationError):
        positive_int("abc")


def test_positive_int_empty_string():
    with pytest.raises(ValidationError):
        positive_int("")


def test_positive_int_none():
    with pytest.raises(ValidationError):
        positive_int(None)


def test_positive_float_valid_float():
    assert positive_float(3.14) == 3.14


def test_positive_float_valid_float_string():
    assert positive_float("3.14") == 3.14


def test_positive_float_valid_int():
    assert positive_float(42) == 42.0


def test_positive_float_valid_int_string():
    assert positive_float("42") == 42.0


def test_positive_float_negative_float():
    with pytest.raises(ValidationError):
        positive_float(-3.14)


def test_positive_float_negative_float_string():
    with pytest.raises(ValidationError):
        positive_float("-3.14")


def test_positive_float_negative_int():
    with pytest.raises(ValidationError):
        positive_float(-42)


def test_positive_float_negative_int_string():
    with pytest.raises(ValidationError):
        positive_float("-42")


def test_positive_float_zero_float():
    with pytest.raises(ValidationError):
        positive_float(0.0)


def test_positive_float_zero_float_string():
    with pytest.raises(ValidationError):
        positive_float("0.0")


def test_positive_float_zero_int():
    with pytest.raises(ValidationError):
        positive_float(0)


def test_positive_float_zero_int_string():
    with pytest.raises(ValidationError):
        positive_float("0")


def test_positive_float_invalid_string():
    with pytest.raises(ValidationError):
        positive_float("abc")


def test_positive_float_empty_string():
    with pytest.raises(ValidationError):
        positive_float("")


def test_positive_float_none():
    with pytest.raises(ValidationError):
        positive_float(None)


def test_non_negative_int_valid():
    assert non_negative_int(42) == 42


def test_non_negative_int_valid_string():
    assert non_negative_int("42") == 42


def test_non_negative_int_valid_float():
    assert non_negative_int(42.0) == 42


def test_non_negative_int_valid_float_string():
    assert non_negative_int("42.0") == 42


def test_non_negative_int_zero_int():
    assert non_negative_int(0) == 0


def test_non_negative_int_zero_int_string():
    assert non_negative_int("0") == 0


def test_non_negative_int_zero_float():
    assert non_negative_int(0.0) == 0


def test_non_negative_int_negative_int():
    with pytest.raises(ValidationError):
        non_negative_int(-42)


def test_non_negative_int_negative_int_string():
    with pytest.raises(ValidationError):
        non_negative_int("-42")


def test_non_negative_int_negative_float():
    with pytest.raises(ValidationError):
        non_negative_int(-42.0)


def test_non_negative_int_negative_float_string():
    with pytest.raises(ValidationError):
        non_negative_int("-42.0")


def test_non_negative_int_invalid_float():
    with pytest.raises(ValidationError):
        non_negative_int(42.5)


def test_non_negative_int_invalid_float_string():
    with pytest.raises(ValidationError):
        non_negative_int("42.5")


def test_non_negative_int_invalid_string():
    with pytest.raises(ValidationError):
        non_negative_int("abc")


def test_non_negative_int_empty_string():
    with pytest.raises(ValidationError):
        non_negative_int("")


def test_non_negative_int_none():
    with pytest.raises(ValidationError):
        non_negative_int(None)


def test_non_negative_float_valid_float():
    assert non_negative_float(3.14) == 3.14


def test_non_negative_float_valid_float_string():
    assert non_negative_float("3.14") == 3.14


def test_non_negative_float_valid_int():
    assert non_negative_float(42) == 42.0


def test_non_negative_float_valid_int_string():
    assert non_negative_float("42") == 42.0


def test_non_negative_float_zero_float():
    assert non_negative_float(0.0) == 0.0


def test_non_negative_float_zero_float_string():
    assert non_negative_float("0.0") == 0.0


def test_non_negative_float_zero_int():
    assert non_negative_float(0) == 0.0


def test_non_negative_float_zero_int_string():
    assert non_negative_float("0") == 0.0


def test_non_negative_float_negative_float():
    with pytest.raises(ValidationError):
        non_negative_float(-3.14)


def test_non_negative_float_negative_float_string():
    with pytest.raises(ValidationError):
        non_negative_float("-3.14")


def test_non_negative_float_negative_int():
    with pytest.raises(ValidationError):
        non_negative_float(-42)


def test_non_negative_float_negative_int_string():
    with pytest.raises(ValidationError):
        non_negative_float("-42")


def test_non_negative_float_invalid_string():
    with pytest.raises(ValidationError):
        non_negative_float("abc")


def test_non_negative_float_empty_string():
    with pytest.raises(ValidationError):
        non_negative_float("")


def test_non_negative_float_none():
    with pytest.raises(ValidationError):
        non_negative_float(None)


def test_valid_filepath_valid_path(fixture_tmp_video_filepath: Path):
    assert valid_filepath(fixture_tmp_video_filepath) == fixture_tmp_video_filepath


def test_valid_filepath_valid_string(fixture_tmp_video_filepath: Path):
    assert valid_filepath(str(fixture_tmp_video_filepath)) == fixture_tmp_video_filepath


def test_valid_filepath_nonexistent_path():
    path = Path("nonexistent") / "file.mp4"
    with pytest.raises(ValidationError):
        valid_filepath(path)


def test_valid_filepath_nonexistent_string():
    path = "nonexistent/file.mp4"
    with pytest.raises(ValidationError):
        valid_filepath(path)


def test_valid_filepath_none():
    with pytest.raises(ValidationError):
        valid_filepath(None)


def test_valid_image_format_valid_image_formats():
    for image_format in C.SUPPORTED_IMAGE_FORMATS:
        assert valid_image_format(image_format) == image_format


def test_valid_image_format_valid_image_formats_uppercase():
    for image_format in C.SUPPORTED_IMAGE_FORMATS:
        assert valid_image_format(image_format.upper()) == image_format


def test_valid_image_format_valid_image_formats_start_with_dot():
    for image_format in C.SUPPORTED_IMAGE_FORMATS:
        assert valid_image_format(f".{image_format}") == image_format


def test_valid_image_format_invalid_image_formats_end_with_dot():
    for image_format in C.SUPPORTED_IMAGE_FORMATS:
        with pytest.raises(ValidationError):
            valid_image_format(f"{image_format}.")


def test_valid_image_format_invalid_image_format():
    with pytest.raises(ValidationError):
        valid_image_format("invalid")


def test_valid_image_format_invalid_empty_string():
    with pytest.raises(ValidationError):
        valid_image_format("")


def test_valid_dir_valid_path(tmp_path: Path):
    assert valid_dir(tmp_path) == tmp_path


def test_valid_dir_valid_string(tmp_path: Path):
    assert valid_dir(str(tmp_path)) == tmp_path


def test_valid_dir_with_dot_as_path_returns_cwd():
    assert valid_dir(Path(".")) == Path().cwd()


def test_valid_dir_with_dot_as_string_returns_cwd():
    assert valid_dir(".") == Path().cwd()


def test_valid_dir_with_two_dots_as_path_returns_parent_dir():
    assert valid_dir(Path("..")) == Path().cwd().parent


def test_valid_dir_with_two_dots_as_string_returns_parent_dir():
    assert valid_dir("..") == Path().cwd().parent


def test_valid_dir_valid_path_with_trailing_slash(tmp_path: Path):
    assert valid_dir(tmp_path / "") == tmp_path


def test_valid_dir_valid_string_with_trailing_slash(tmp_path: Path):
    assert valid_dir(str(tmp_path / "")) == tmp_path


def test_valid_dir_invalid_path(tmp_path: Path):
    with pytest.raises(ValidationError):
        valid_dir(tmp_path / "invalid")


def test_valid_dir_invalid_string(tmp_path: Path):
    with pytest.raises(ValidationError):
        valid_dir(str(tmp_path / "invalid"))


def test_valid_dir_with_forward_slash():
    with pytest.raises(ValidationError):
        valid_dir("/")


def test_valid_dir_none():
    with pytest.raises(ValidationError):
        valid_dir(None)


@pytest.mark.parametrize(
    ("start", "stop", "duration", "expected"),
    [
        (0, 1, 2, (0, 1, 2)),  # start < stop < duration
        (1, 3, 3, (1, 3, 3)),  # stop < stop, stop == duration
    ],
)
def test_valid_extraction_range_valid_range(
    start: float, stop: float, duration: float, expected: tuple[float, float, float]
):
    assert valid_extraction_range(start, stop, duration) == expected


@pytest.mark.parametrize(
    ("start", "stop", "duration"),
    [
        (1, 0, 0),  # start > duration
        (1, 1, 2),  # start == stop
        (1, 0, 2),  # start > stop
    ],
)
def test_valid_extraction_range_invalid_range(
    start: float, stop: float, duration: float
):
    with pytest.raises(ValidationError):
        valid_extraction_range(start, stop, duration)


@pytest.mark.parametrize(
    ("filename", "expected_filename"),
    [
        ("filename", "filename"),
        ("filename (1)", "filename (1)"),
        ("filename-with-dashes", "filename-with-dashes"),
        ("filename with spaces", "filename with spaces"),
        ("filename_with_underscores", "filename_with_underscores"),
        ("filenamewithsuffix.mp4", "filenamewithsuffix.mp4"),
        ("filenameendswithdot.", "filenameendswithdot."),
        (".filenamestartswithdot", ".filenamestartswithdot"),
        (" filename starts with space", " filename starts with space"),
        ("filename ends with space ", "filename ends with space "),
        (
            "filename_with_symbols!@#$%^&()_+{}[];'",
            "filename_with_symbols!@#$%^&()_+{}[];'",
        ),
        (".mp4.", ".mp4."),
        (".mp4", ".mp4"),
        (".", "."),
        (" ", " "),
    ],
)
def test_valid_filename_valid_filenames(filename: str, expected_filename: str):
    assert valid_filename(filename) == expected_filename


@pytest.mark.parametrize(
    ("filename"),
    [
        ("file|with|invalid|characters.mp4"),
        ("file\\\\with\\\\backslashes.mp4"),
        ("file/with/forward/slashes.mp4"),
        ("file*with*asterisks.mp4"),
        ("file?with?question?marks.mp4"),
        ("file<with<less<than.mp4"),
        ("file>with>greater>than.mp4"),
        ("file:with:colons.mp4"),
        ('file"with"quotes.mp4'),
        ("file|with|pipes.mp4"),
        (""),
        (None),
    ],
)
def test_valid_filename_invalid_filenames(filename: str):
    with pytest.raises(ValidationError):
        valid_filename(filename)


@pytest.mark.parametrize(
    ("timestamp", "expected"),
    [
        ("0:00", "0:00"),
        ("00:00", "00:00"),
        ("0:00:00", "0:00:00"),
        ("00:00:00", "00:00:00"),
        ("0:00:00.0", "0:00:00"),
        ("0:00:00.01", "0:00:00"),
        ("0:00:00.99", "0:00:00"),
        ("12:34:56", "12:34:56"),
        ("1:23:45", "1:23:45"),
        ("12:34", "12:34"),
        ("1:23", "1:23"),
        ("59:59:59", "59:59:59"),
    ],
)
def test_valid_timestamp_valid(timestamp: str, expected: str):
    assert valid_timestamp(timestamp) == expected


@pytest.mark.parametrize(
    ("timestamp"),
    [
        (""),
        (" "),
        (":"),
        ("0"),
        ("00"),
        (":0"),
        (":00"),
        ("0:"),
        ("00:"),
        ("0:0"),
        ("00:0"),
        ("0:0:0"),
        ("0:0:00"),
        ("0:00:0"),
        ("00:0:0"),
        ("00:00:0"),
        ("60:00:00"),
        ("00:60:00"),
        ("00:00:60"),
        ("60:60:60"),
        ("abc"),
        ("ab:cd:ef"),
        ("x0:00:00"),
        ("0x:00:00"),
        ("00:x0:00"),
        ("00:0x:00"),
        ("00:00:x0"),
        ("00:00:0x"),
        (None),
    ],
)
def test_valid_timestamp_invalid(timestamp: str):
    with pytest.raises(ValidationError):
        valid_timestamp(timestamp)


def test_valid_audio_format_valid_audio_formats():
    for audio_format in C.SUPPORTED_AUDIO_FORMATS:
        assert valid_audio_format(audio_format) == audio_format


def test_valid_audio_format_valid_audio_formats_uppercase():
    for audio_format in C.SUPPORTED_AUDIO_FORMATS:
        assert valid_audio_format(audio_format.upper()) == audio_format


def test_valid_audio_format_valid_audio_formats_start_with_dot():
    for audio_format in C.SUPPORTED_AUDIO_FORMATS:
        assert valid_audio_format(f".{audio_format}") == audio_format


def test_valid_audio_format_invalid_audio_formats_end_with_dot():
    for audio_format in C.SUPPORTED_AUDIO_FORMATS:
        with pytest.raises(ValidationError):
            valid_audio_format(f"{audio_format}.")


def test_valid_audio_format_invalid_audio_format():
    with pytest.raises(ValidationError):
        valid_audio_format("invalid")


def test_valid_audio_format_invalid_empty_string():
    with pytest.raises(ValidationError):
        valid_audio_format("")


@pytest.mark.parametrize(
    ("dimensions", "expected"),
    [
        ((1, 2), (1, 2)),  # width < height
        ((2, 1), (2, 1)),  # width > height
        ((1, 1), (1, 1)),  # width == height
        ((1.0, 1.0), (1, 1)),  # whole float
        (("1", "1"), (1, 1)),  # integer as string
        (("1.0", "1.0"), (1, 1)),  # whole float as string
    ],
)
def test_valid_dimensions_valid_dimensions(
    dimensions: tuple[int, int], expected: tuple[int, int]
):
    assert valid_dimensions(dimensions) == expected


@pytest.mark.parametrize(
    ("dimensions"),
    [
        ((-1, -1)),  # width or height cannot be negative
        ((0, 0)),  # width or height cannot be (0)
        ((1, 2, 3)),  # dimensions cannot have more than two values
        ((1,)),  # dimensions cannot have less than two values
        (()),  # dimensions cannot be empty
        ((1.5, 1.5)),  # dimensions cannot be floating point values
        (("1.5", "1.5")),  # dimensions cannot be floating point values as strings
        (("abc", "abc")),  # dimensions cannot be strings
    ],
)
def test_valid_dimensions_invalid_dimensions(dimensions: tuple[int, int]):
    with pytest.raises(ValidationError):
        valid_dimensions(dimensions)


def test_valid_rotate_value_valid_rotate_ints():
    for rotate_value in C.VALID_ROTATE_VALUES:
        assert valid_rotate_value(rotate_value) == rotate_value


def test_valid_rotate_value_valid_rotate_floats():
    for rotate_value in C.VALID_ROTATE_VALUES:
        assert valid_rotate_value(float(rotate_value)) == rotate_value


def test_valid_rotate_value_valid_rotate_int_strings():
    for rotate_value in C.VALID_ROTATE_VALUES:
        assert valid_rotate_value(str(rotate_value)) == rotate_value


@pytest.mark.parametrize(
    ("rotate_value"),
    [
        ("0.0"),
        ("90.0"),
        ("180.0"),
        ("270.0"),
        (-1),
        ("-1"),
        (-1.0),
        ("-1.0"),
        (42),
        ("42"),
        (42.0),
        ("42.0"),
        ("abc"),
        (""),
    ],
)
def test_valid_rotate_value_invalid_rotate_values(rotate_value: float | int | str):
    with pytest.raises(ValidationError):
        valid_rotate_value(rotate_value)


def test_valid_start_time_with_valid_start_time_strings():
    assert valid_start_time("0") == 0.0
    assert valid_start_time("60") == 60.0
    assert valid_start_time("0:00") == "0:00"
    assert valid_start_time("0:00:00") == "0:00:00"
    assert valid_start_time("0:00:00.0") == "0:00:00"
    assert valid_start_time("0:00:00.9") == "0:00:00"


def test_valid_start_time_with_valid_start_time_ints():
    assert valid_start_time(0) == 0
    assert valid_start_time(1) == 1
    assert valid_start_time(60) == 60
    assert valid_start_time(3600) == 3600


def test_valid_start_time_with_valid_start_time_floats():
    assert valid_start_time(0.0) == 0.0
    assert valid_start_time(0.9) == 0.9
    assert valid_start_time(1.0) == 1.0
    assert valid_start_time(1.9) == 1.9
    assert valid_start_time(60.0) == 60.0
    assert valid_start_time(60.9) == 60.9
    assert valid_start_time(3600.0) == 3600.0
    assert valid_start_time(3600.9) == 3600.9


def test_valid_stop_time_with_valid_stop_time_strings():
    assert valid_stop_time("1") == 1.0
    assert valid_stop_time("60") == 60.0
    assert valid_stop_time("1:00") == "1:00"
    assert valid_stop_time("1:00:00") == "1:00:00"
    assert valid_stop_time("1:00:00.0") == "1:00:00"
    assert valid_stop_time("1:00:00.9") == "1:00:00"


def test_valid_stop_time_with_valid_stop_time_ints():
    assert valid_stop_time(1) == 1
    assert valid_stop_time(60) == 60
    assert valid_stop_time(3600) == 3600


def test_valid_stop_time_with_valid_stop_time_floats():
    assert valid_stop_time(0.01) == 0.01
    assert valid_stop_time(1.0) == 1.0
    assert valid_stop_time(1.9) == 1.9
    assert valid_stop_time(60.0) == 60.0
    assert valid_stop_time(60.9) == 60.9
    assert valid_stop_time(3600.0) == 3600.0
    assert valid_stop_time(3600.9) == 3600.9


def test_valid_video_filepath_with_supported_existing_filepath(
    fixture_tmp_video_filepath,
):
    assert (
        valid_filepath(fixture_tmp_video_filepath, is_video=True)
        == fixture_tmp_video_filepath
    )


def test_valid_video_filepath_with_supported_existing_filepath_string(
    fixture_tmp_video_filepath,
):
    path_str = str(fixture_tmp_video_filepath)
    assert valid_filepath(path_str, is_video=True) == fixture_tmp_video_filepath


def test_valid_video_filepath_with_supported_existing_filepath_lowercase_suffix(
    fixture_tmp_video_filepath,
):
    path = fixture_tmp_video_filepath.with_suffix(".mp4")
    assert valid_filepath(path, is_video=True) == fixture_tmp_video_filepath


def test_valid_video_filepath_with_supported_existing_filepath_mixedcase_suffix(
    fixture_tmp_video_filepath,
):
    path = fixture_tmp_video_filepath.with_suffix(".Mp4")
    assert valid_filepath(path, is_video=True) == fixture_tmp_video_filepath


def test_valid_video_filepath_with_supported_existing_filepath_uppercase_suffix(
    fixture_tmp_video_filepath,
):
    path = fixture_tmp_video_filepath.with_suffix(".MP4")
    assert valid_filepath(path, is_video=True) == fixture_tmp_video_filepath


def test_valid_video_filepath_with_supported_nonexistant_video_filepath(tmp_path: Path):
    path = tmp_path / "t.mp4"
    with pytest.raises(ValidationError):
        valid_filepath(path, is_video=True)


def test_valid_video_filepath_with_supported_nonexistant_video_filepath_string():
    with pytest.raises(ValidationError):
        valid_filepath("t.mp4", is_video=True)


def test_valid_video_filepath_with_unsupported_existing_file_suffix(
    fixture_tmp_text_filepath,
):
    with pytest.raises(ValidationError):
        valid_filepath(fixture_tmp_text_filepath, is_video=True)


def test_valid_video_filepath_with_unsupported_nonexistant_file_suffix():
    with pytest.raises(ValidationError):
        valid_filepath("t.txt", is_video=True)


def test_valid_video_filepath_with_directory_path(tmp_path: Path):
    with pytest.raises(ValidationError):
        valid_filepath(tmp_path, is_video=True)


def test_valid_video_file_suffix_with_supported_video_file_suffixes():
    for suffix in C.SUPPORTED_VIDEO_FORMATS:
        assert valid_video_file_suffix(suffix) == suffix


def test_valid_video_file_suffix_with_unsupported_video_file_suffix():
    with pytest.raises(ValidationError):
        valid_video_file_suffix(".abc")


def test_valid_volume_with_valid_volume():
    assert valid_volume(-1) == 0.0
    assert valid_volume(0.0) == 0.0
    assert valid_volume(0.5) == 0.5
    assert valid_volume(1.0) == 1.0
    assert valid_volume(1.5) == 1.5
    assert valid_volume("2") == 2.0
    assert valid_volume("3.0") == 3.0


def test_valid_volume_with_invalid_input():
    with pytest.raises(ValidationError):
        valid_volume("abc")
    with pytest.raises(ValidationError):
        valid_volume(None)
    with pytest.raises(ValidationError):
        valid_volume("")
    with pytest.raises(ValidationError):
        valid_volume(" ")



"""Containers for extraction results."""
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from videoxt.utils import ToJsonMixin


@dataclass
class Result(ToJsonMixin):
    """
    A container for the result of an extraction.

    Attributes:
    -----
        `success` (Optional[bool]):
            Whether the extraction was successful.
        `method` (Optional[str]):
            The name of the method that was called.
        `message` (Optional[str]):
            A message describing the result of the extraction.
        `destpath` (Optional[Path]):
            The path to the destination file or directory.
        `elapsed_time` (Optional[float]):
            The number of seconds it took to complete the extraction.

    Methods:
    -----
        - `json()` -> `str`: Return a JSON string representation of the request.
    """

    success: Optional[bool] = None
    method: Optional[str] = None
    message: Optional[str] = None
    destpath: Optional[Path] = None
    elapsed_time: Optional[float] = None



"""Contains custom exceptions raised by videoxt."""


class VideoXTError(Exception):
    """A videoxt specific error occurred."""

    pass


class InvalidExtractionMethod(VideoXTError):
    """An invalid extraction method was requested."""

    pass


class ClosedVideoCaptureError(VideoXTError):
    """A `cv2.VideoCapture` was unexpectedly closed."""

    pass


class VideoValidationError(VideoXTError):
    """A video file failed a validation step."""

    pass


class PreparationError(VideoXTError):
    """An error occurred while preparing a request."""

    pass


class ValidationError(VideoXTError):
    """A user input validation step failed."""

    pass


class NoAudioError(VideoXTError):
    """A video has no audio and it is required for the operation."""

    pass


class FrameReadError(VideoXTError):
    """An error occurred while reading a frame from a video using `cv2`."""

    pass


class VideoCaptureSetError(VideoXTError):
    """An error occurred while setting a VideoCapture property using `cv2`."""

    pass


class FrameWriteError(VideoXTError):
    """An error occurred while writing a frame to file using `cv2`."""

    pass


class GifWriteError(VideoXTError):
    """An error occurred while writing a gif to file."""

    pass


class AudioWriteError(VideoXTError):
    """An error occurred while writing an audio file."""

    pass


class ClipWriteError(VideoXTError):
    """An error occurred while writing a clip to file."""

    pass


class BuildImagePathError(VideoXTError):
    """An error occurred while building an image path."""

    pass



"""
/* VideoXT */

A simple library and CLI tool for extracting audio, individual frames, short clips and
GIFs from videos.

Contents:
-----
- `extract`: Extract audio, individual frames, short clips and GIFs from videos.
- `extract_audio`: Extract audio from a video file.
- `extract_clip`: Extract a short clip from a video file as `mp4`.
- `extract_frames`: Extract individual frames from a video and save them as images.
- `extract_gif`: Create a GIF from a video between two specified points.

Basic Usage:
-----
>>> import videoxt as vxt
>>> video_filepath = "path/to/video.mp4"  # or pathlib.Path("path/to/video.mp4")

# Extract all audio.
>>> vxt.extract("audio", video_filepath)  # or...
>>> vxt.extract_audio(video_filepath)

# Save every 30th frame of the video to disk as an image.
>>> vxt.extract_frames(
...     video_filepath,
...     capture_rate=30
... )

# Extract a clip between 12:29 and 12:34 from the video.
>>> vxt.extract_clip(
...     video_filepath,
...     start_time="12:29",
...     stop_time="12:34"
... )

# Extract a GIF of the first 2 seconds of the video that bounces back and forth.
>>> vxt.extract_gif(
...     video_filepath,
...     stop_time=2,
...     bounce=True
... )
"""

# flake8: noqa

from videoxt.api import (
    extract,
    extract_audio,
    extract_clip,
    extract_frames,
    extract_gif,
)



"""Utility functions and classes used throughout the library."""
import json
from collections import defaultdict
from dataclasses import asdict, dataclass
from datetime import timedelta
from pathlib import Path
from typing import Any, ClassVar, Optional, Protocol

from rich import print

from videoxt.constants import ExtractionMethod
from videoxt.validators import positive_float, positive_int, valid_filename


def timestamp_to_seconds(timestamp: str) -> float:
    """
    Convert a timestamp string to the total number of seconds (float) it represents.

    Accepts the formats "HH:MM:SS", "H:MM:SS", "MM:SS", "M:SS", "SS" or "S".
    Microseconds are truncated.

    Usage:
    -----

        >>> timestamp_to_seconds("1:01")
        61.0
        >>> timestamp_to_seconds("1:01.987654321")
        61.0
        >>> timestamp_to_seconds("10:00")
        600.0
        >>> timestamp_to_seconds("1:00:00")
        3600.0

    Args:
    -----
        `timestamp` (str):
            A timestamp string in the format "HH:MM:SS", "H:MM:SS", "MM:SS", "M:SS",
            "SS" or "S"

    Returns:
    -----
        `float`: The number of seconds converted from the timestamp string.
    """
    timestamp = timestamp.split(".")[0]
    time_parts = timestamp.split(":")
    total_seconds = sum(
        float(part) * 60**exponent
        for exponent, part in enumerate(reversed(time_parts))
    )

    return total_seconds


def seconds_to_timestamp(seconds: float) -> str:
    """
    Convert seconds to a time duration string in the format "HH:MM:SS" or "H:MM:SS".

    Microseconds are truncated.

    Usage:
    -----
        >>> seconds_to_timestamp(60)
        '0:01:00'
        >>> seconds_to_timestamp(60.987654321)
        '0:01:00'
        >>> seconds_to_timestamp(600)
        '0:10:00'
        >>> seconds_to_timestamp(3600)
        '1:00:00'

    Args:
    -----
        `seconds` (float): The number of seconds to convert to a timestamp.

    Returns:
    -----
        `str`: A timestamp string in the format "HH:MM:SS" or "H:MM:SS".
    """
    if seconds <= 0:
        return "0:00:00"

    return str(timedelta(seconds=int(seconds)))


def timedelta_to_timestamp(duration: timedelta) -> str:
    """
    Convert a timedelta to a time duration string in the format "HH:MM:SS".

    Microseconds are truncated.

    Usage:
    -----
        >>> timedelta_to_timestamp(timedelta(seconds=59))
        '00:00:59'
        >>> timedelta_to_timestamp(timedelta(seconds=61.987654321))
        '00:01:01'
        >>> timedelta_to_timestamp(timedelta(seconds=600))
        '00:10:00'

    Args:
    -----
        `duration` (datetime.timedelta): The duration to convert to a timestamp.

    Returns:
    -----
        `str`: A timestamp string in the format "HH:MM:SS".
    """
    total_seconds = duration.total_seconds()

    if total_seconds < 0:
        raise ValueError(
            f"Invalid duration: timedelta must be non-negative: {duration}"
        )

    hours, remainder = divmod(total_seconds, 3600)
    minutes, seconds = divmod(remainder, 60)

    return f"{int(hours):02}:{int(minutes):02}:{int(seconds):02}"


def append_enumeration(index: int, tag: Optional[str] = None) -> str:
    """
    Return an enumerated file name with the given index and optional tag.

    The returned string is intended to be appended to a file name or directory name.
    If the tag is None, only the index is used to enumerate the file name starting
    from 1. If a tag is provided, the tag is appended to the file name before the
    index. Mimics the behavior of Windows when creating a file or directory with the
    same name as an existing file or directory.

    Usage:
    -----
        >>> append_enumeration(1)
        ' (1)'
        >>> append_enumeration(2)
        ' (2)'
        >>> append_enumeration(1, tag="_vxt")
        '_vxt'
        >>> append_enumeration(2, tag="_vxt")
        '_vxt (2)'

    Args:
    -----
        `index` (int):
            The enumeration index. Must be greater than 0.
        `tag` (Optional[str]):
            The tag to enumerate. If None, only the index is used.

    Returns:
    -----
        `str`: The enumerated string to append to a file name or directory name.
    """
    if index < 1:
        raise ValueError(f"Enumeration index must be greater than 0, got {index}.")

    if tag is None:
        return f" ({index})"

    # Ensure the tag doesn't contain invalid characters.
    tag = valid_filename(tag)

    return f"{tag} ({index})" if index > 1 else tag


def enumerate_dir(directory: Path, tag: Optional[str] = None) -> Path:
    """
    Return a non-existent, potentially enumerated directory path.

    See `videoxt.utils.append_enumeration` for more information on how the directory
    name is enumerated with or without a tag.

    Usage:
    -----
        >>> from pathlib import Path
        >>> path = Path("test_dir")  # Assume 'test_dir' doesn't exist.
        >>> enumerate_dir(path, tag="_frames")
        Path('test_dir')
        >>> enumerate_dir(path, tag="_frames")
        Path('test_dir_frames')
        >>> enumerate_dir(path, tag="_frames")
        Path('test_dir_frames (2)')

    Args:
    -----
        `directory`:
            The path to the directory to potentially enumerate.
        `tag` (Optional[str]):
            The tag to enumerate. If None, only the index is used.

    Returns:
    -----
        `pathlib.Path`: The path to a non-existent directory.
    """
    if not directory.exists():
        return directory

    index = 1
    while True:
        append_str = append_enumeration(index, tag=tag)
        new_dir = directory.with_name(f"{directory.name}{append_str}")
        if not new_dir.exists():
            return new_dir
        index += 1


def enumerate_filepath(filepath: Path, tag: Optional[str] = None) -> Path:
    """
    Return a non-existent, potentially enumerated file path.

    See `videoxt.utils.append_enumeration` for more information on how the file name
    is enumerated with or without a tag.

    Usage:
    -----
        >>> from pathlib import Path
        >>> filepath = Path('test.mp4')  # Assume 'test.mp4' doesn't exist.
        >>> enumerate_filepath(filepath, tag="_vxt")
        Path('test.mp4')
        >>> enumerate_filepath(filepath, tag="_vxt")
        Path('test_vxt.mp4')
        >>> enumerate_filepath(filepath, tag="_vxt")
        Path('test_vxt (2).mp4')

    Args:
    -----
        `filepath` (pathlib.Path): The path to the file to potentially enumerate.

    Returns:
    -----
        `pathlib.Path`: The path to a non-existent file.
    """
    if not filepath.exists():
        return filepath

    index = 1
    while True:
        append_str = append_enumeration(index, tag)
        new_filename = f"{filepath.stem}{append_str}{filepath.suffix}"
        new_filepath = filepath.with_name(new_filename)

        if not new_filepath.exists():
            return new_filepath

        index += 1


def calculate_duration(frame_count: int, fps: float) -> timedelta:
    """
    Return a timedelta representing the duration of a video using frame count and fps.

    Args:
    -----
        `frame_count` (int):
            The total number of frames in the video.
        `fps` (float):
            The frames per second of the video.

    Returns:
    -----
        `datetime.timedelta`: The duration of the video.

    Raises:
    -----
        `ZeroDivisionError`: If the fps is 0.
    """
    frame_count = positive_int(frame_count)
    fps = positive_float(fps)

    try:
        td = timedelta(seconds=frame_count / fps)
    except ZeroDivisionError:
        raise ZeroDivisionError("FPS cannot be 0.")
    else:
        return td


def convert_bytes(n: int) -> str:
    """
    Convert n bytes (int) to a human readable string.

    Usage:
    -----
        >>> convert_bytes(1)
        '1.00 bytes'
        >>> convert_bytes(1024)
        '1.00 KB'
        >>> convert_bytes(1024**2)
        '1.00 MB'
        >>> convert_bytes(1024**3)
        '1.00 GB'
        >>> convert_bytes(1024**4)
        '1.00 TB'
        >>> convert_bytes(1024**5)
        '1.00 PB'

    Args:
    -----
        `n` (int): The number of bytes to convert.

    Returns:
    -----
        `str`: A human readable string representing the number of bytes.
    """
    n = positive_int(n)
    for size in ["bytes", "KB", "MB", "GB", "TB"]:
        if n < 1024.0:
            return f"{n:.2f} {size}"
        n /= 1024.0

    return f"{n:.2f} PB"


class CustomJSONEncoder(json.JSONEncoder):
    """A custom JSON encoder for types that are not JSON serializable."""

    def default(self, obj: Any) -> Any:
        """Return a JSON serializable representation of the object."""
        if isinstance(obj, Path):
            return obj.resolve().as_posix()
        if isinstance(obj, (Path, timedelta)):
            return str(obj)
        if isinstance(obj, ExtractionMethod):
            return obj.value
        return super().default(obj)


@dataclass
class ToJsonMixin:
    """A mixin for dataclasses that can be represented as JSON."""

    def json(self, skip_private_keys: bool = False) -> str:
        """
        Return a JSON string representation of the dataclass.

        Args:
        -----
            `skip_private_keys` (bool):
                If True, private keys (keys starting with '_') are not included in the
                JSON string.

        Returns:
        -----
            `str`: JSON string representation of the dataclass.
        """
        d = remove_private_keys(asdict(self)) if skip_private_keys else asdict(self)
        return json.dumps(d, indent=2, cls=CustomJSONEncoder)

    def verbose_print(self, title: str) -> None:
        """Print the public keys of the JSON to console with a title."""
        color_map = {
            "PreparedRequest": "bright_magenta",
            "Result": "yellow",
        }
        color = color_map.get(title, "white")
        print(f"<[{color}]{title}[/{color}]>\n{self.json(skip_private_keys=True)}")


class DataclassType(Protocol):
    """Protocol representing dataclass attributes for type-hinting purposes."""

    __dataclass_fields__: ClassVar[dict]


def remove_private_keys(d: dict[str, Any]) -> dict[str, Any]:
    """
    Return a copy of the dictionary without private keys (keys starting with '_').

    Args:
    -----
        `d` (dict): The dictionary you want to remove private keys from.

    Returns:
    -----
        `dict[str, Any]`: A new dictionary containing public keys only.
    """
    copied_dict: defaultdict[str, Any] = defaultdict(dict)
    for k, v in d.items():
        if not k.startswith("_"):
            copied_dict[k] = remove_private_keys(v) if isinstance(v, dict) else v
    return dict(copied_dict)


def parse_kwargs(kwargs: dict[str, Any], obj: DataclassType) -> dict[str, Any]:
    """
    Return the keys and values in `kwargs` present in the `obj` dataclass attributes.

    Usage:
    -----
        >>> from dataclasses import dataclass
        >>> @dataclass
        ... class Test:
        ...     a: int
        ...     b: int
        >>> parse_kwargs({"a": 1, "b": 2, "c": 3}, Test)
        {'a': 1, 'b': 2}

    Args:
    -----
        `kwargs` (dict[str, Any]):
            A dictionary to filter down to keys present in the `obj` attributes.
        `obj` (dataclasses.dataclass):
            A dataclass to filter the `kwargs` dictionary with.

    Returns:
    -----
        `dict`: A dictionary of keyword arguments present in the `obj` attributes.
    """
    return {k: v for k, v in kwargs.items() if k in obj.__dataclass_fields__}



import json
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path

import pytest

from videoxt.constants import ExtractionMethod
from videoxt.exceptions import ValidationError
from videoxt.utils import (
    CustomJSONEncoder,
    ToJsonMixin,
    append_enumeration,
    calculate_duration,
    convert_bytes,
    enumerate_dir,
    enumerate_filepath,
    parse_kwargs,
    seconds_to_timestamp,
    timedelta_to_timestamp,
    timestamp_to_seconds,
)


def test_timestamp_to_seconds_valid_input():
    """
    Test that the function returns a valid number of seconds for a given timestamp.
    """
    assert timestamp_to_seconds("0:00") == 0
    assert timestamp_to_seconds("0:01") == 1
    assert timestamp_to_seconds("0:59") == 59
    assert timestamp_to_seconds("1:00") == 60
    assert timestamp_to_seconds("1:01") == 61
    assert timestamp_to_seconds("0:00:00") == 0
    assert timestamp_to_seconds("0:00:01") == 1
    assert timestamp_to_seconds("0:00:59") == 59
    assert timestamp_to_seconds("0:01:00") == 60
    assert timestamp_to_seconds("0:01:01") == 61
    assert timestamp_to_seconds("1:00:00") == 3600
    assert timestamp_to_seconds("1:00:01") == 3601
    assert timestamp_to_seconds("1:00:59") == 3659
    assert timestamp_to_seconds("1:01:00") == 3660
    assert timestamp_to_seconds("1:01:01") == 3661


def test_timestamp_to_seconds_abnormal_but_valid_input():
    """
    Test that the function returns a valid number of seconds for an abnormal timestamp.

    The following timestamps are not valid according to the function's docstring, but
    they should be treated be handled by the function.
    """
    assert timestamp_to_seconds("0") == 0
    assert timestamp_to_seconds("1") == 1
    assert timestamp_to_seconds("1.1") == 1
    assert timestamp_to_seconds("1.1.1") == 1
    assert timestamp_to_seconds("59") == 59
    assert timestamp_to_seconds("60") == 60
    assert timestamp_to_seconds("61") == 61
    assert timestamp_to_seconds("61.1") == 61
    assert timestamp_to_seconds("61.1.1") == 61
    assert timestamp_to_seconds("1:01.1") == 61
    assert timestamp_to_seconds("1:01.1.1") == 61
    assert timestamp_to_seconds("0:01:01.1") == 61
    assert timestamp_to_seconds("0:01:01.1.1") == 61
    assert timestamp_to_seconds("1:01:01.1") == 3661
    assert timestamp_to_seconds("1:01:01.1.1") == 3661
    assert timestamp_to_seconds("1:1") == 61
    assert timestamp_to_seconds("1:1:1") == 3661
    assert timestamp_to_seconds("1:1:1.1") == 3661
    assert timestamp_to_seconds("1:1:1.1.1") == 3661


def test_timestamp_to_seconds_invalid_input():
    """
    Test that the function raises a ValueError if the input is not a valid timestamp.
    """
    with pytest.raises(ValueError):
        timestamp_to_seconds("invalid")


def test_timestamp_to_seconds_empty_string():
    """Test that the function raises a ValueError if the input is an empty string."""
    with pytest.raises(ValueError):
        timestamp_to_seconds("")


def test_seconds_to_timestamp_valid_input():
    """
    Test that the function returns a valid timestamp string for a given number of
    seconds.
    """
    assert seconds_to_timestamp(-1) == "0:00:00"
    assert seconds_to_timestamp(0) == "0:00:00"
    assert seconds_to_timestamp(0.0) == "0:00:00"
    assert seconds_to_timestamp(0.001) == "0:00:00"
    assert seconds_to_timestamp(0.999) == "0:00:00"
    assert seconds_to_timestamp(1) == "0:00:01"
    assert seconds_to_timestamp(1.0) == "0:00:01"
    assert seconds_to_timestamp(1.001) == "0:00:01"
    assert seconds_to_timestamp(1.999) == "0:00:01"
    assert seconds_to_timestamp(59) == "0:00:59"
    assert seconds_to_timestamp(59.0) == "0:00:59"
    assert seconds_to_timestamp(59.001) == "0:00:59"
    assert seconds_to_timestamp(59.999) == "0:00:59"
    assert seconds_to_timestamp(60) == "0:01:00"
    assert seconds_to_timestamp(60.0) == "0:01:00"
    assert seconds_to_timestamp(60.001) == "0:01:00"
    assert seconds_to_timestamp(60.999) == "0:01:00"
    assert seconds_to_timestamp(61) == "0:01:01"
    assert seconds_to_timestamp(61.0) == "0:01:01"
    assert seconds_to_timestamp(61.001) == "0:01:01"
    assert seconds_to_timestamp(61.999) == "0:01:01"
    assert seconds_to_timestamp(3600) == "1:00:00"
    assert seconds_to_timestamp(3600.0) == "1:00:00"
    assert seconds_to_timestamp(3600.001) == "1:00:00"
    assert seconds_to_timestamp(3600.999) == "1:00:00"
    assert seconds_to_timestamp(3601) == "1:00:01"
    assert seconds_to_timestamp(3601.0) == "1:00:01"
    assert seconds_to_timestamp(3601.001) == "1:00:01"
    assert seconds_to_timestamp(3601.999) == "1:00:01"
    assert seconds_to_timestamp(3659) == "1:00:59"
    assert seconds_to_timestamp(3659.0) == "1:00:59"
    assert seconds_to_timestamp(3659.001) == "1:00:59"
    assert seconds_to_timestamp(3659.999) == "1:00:59"
    assert seconds_to_timestamp(3661) == "1:01:01"
    assert seconds_to_timestamp(3661.0) == "1:01:01"
    assert seconds_to_timestamp(3661.001) == "1:01:01"
    assert seconds_to_timestamp(3661.999) == "1:01:01"


def test_seconds_to_timestamp_invalid_input():
    """Test that the function raises a TypeError if the input is not a numeric value."""
    with pytest.raises(TypeError):
        seconds_to_timestamp("invalid")


def test_seconds_to_timestamp_empty_string():
    """Test that the function raises a ValueError if the input is an empty string."""
    with pytest.raises(TypeError):
        seconds_to_timestamp("")


def test_enumerate_dir_existing_dir(tmp_path: Path):
    """Test that the directory path is enumerated if the directory already exists.

    `tmp_path` is a built-in pytest fixture that creates a temporary directory.
    """
    expected_path = tmp_path.with_name(f"{tmp_path.name}_vxt")
    assert enumerate_dir(tmp_path, tag="_vxt") == expected_path


def test_enumerate_dir_new_dir(tmp_path: Path):
    """Test that the same directory path is returned if the directory does not exist.

    `tmp_path` is a built-in pytest fixture that creates a temporary directory.
    """
    expected_path = tmp_path / "new_dir"
    assert enumerate_dir(expected_path) == expected_path


def test_enumerate_filepath_existing_file(fixture_tmp_text_filepath: Path):
    """Test that the filepath is enumerated if the file already exists."""
    stem = fixture_tmp_text_filepath.stem
    suffix = fixture_tmp_text_filepath.suffix
    expected_path = fixture_tmp_text_filepath.with_name(f"{stem}_vxt{suffix}")
    assert enumerate_filepath(fixture_tmp_text_filepath, tag="_vxt") == expected_path


def test_enumerate_filepath_new_file(fixture_tmp_text_filepath: Path):
    """Test that the same filepath is returned if the file does not exist."""
    expected_path = fixture_tmp_text_filepath.with_name("new_file.txt")
    assert enumerate_filepath(expected_path) == expected_path


@dataclass
class Image:
    """Test dataclass for the `utils.parse_kwargs` function."""

    filename: str
    width: int
    height: int


@pytest.mark.parametrize(
    "kwargs, expected",
    [
        # Test case 1: all kwargs are defined as fields in the `Image` dataclass.
        (
            {
                "filename": "image.jpg",
                "width": 1920,
                "height": 1080,
            },
            {"filename": "image.jpg", "width": 1920, "height": 1080},
        ),
        # Test case 2: all kwargs are defined as fields in the `Image` dataclass except
        # `extra_field`.
        (
            {
                "filename": "image.jpg",
                "width": 1920,
                "height": 1080,
                "extra_field": "",
            },
            {"filename": "image.jpg", "width": 1920, "height": 1080},
        ),
        # Test case 3: select kwargs are defined as fields in the `Image` dataclass.
        (
            {"width": 1920, "height": 1080},
            {"width": 1920, "height": 1080},
        ),
        # Test case 4: one of two kwargs are defined as fields in the `Image` dataclass.
        (
            {"filename": "image.jpg", "extra_field": ""},
            {"filename": "image.jpg"},
        ),
        # Test case 5: one of one kwarg is defined as a field in the `Image` dataclass.
        (
            {"filename": "image.jpg"},
            {"filename": "image.jpg"},
        ),
        # Test case 6: no kwargs are defined as fields in the `Image` dataclass.
        (
            {"extra_field": "", "another_extra_field": ""},
            {},
        ),
        # Test case 7: one of one kwarg is not defined as a field in the `Image`
        # dataclass.
        (
            {"extra_field": ""},
            {},
        ),
        # Test case 8: empty kwargs passed.
        (
            {},
            {},
        ),
        # Test case 9: all kwargs are defined as fields in the `Image` dataclass, but
        # one is None.
        (
            {
                "filename": "image.jpg",
                "width": 1920,
                "height": None,
            },
            {"filename": "image.jpg", "width": 1920, "height": None},
        ),
    ],
)
def test_parse_kwargs(kwargs: dict, expected: dict):
    """Test that the `parse_kwargs` function returns a dictonary containing only the
    keyword arguments that match the fields in a given dataclass."""
    assert parse_kwargs(kwargs, Image) == expected


def test_append_enumeration_when_only_index_is_entered():
    assert append_enumeration(1) == " (1)"
    assert append_enumeration(2) == " (2)"
    assert append_enumeration(3) == " (3)"


def test_append_enumeration_when_index_and_tag_are_entered():
    assert append_enumeration(1, tag="_vxt") == "_vxt"
    assert append_enumeration(2, tag="_vxt") == "_vxt (2)"
    assert append_enumeration(3, tag="_vxt") == "_vxt (3)"


def test_append_enumeration_when_index_is_0():
    with pytest.raises(ValueError):
        append_enumeration(0)


def test_calculate_duration_with_valid_inputs():
    frame_count = 120
    fps = 30.0
    expected_duration = timedelta(seconds=frame_count / fps)
    result = calculate_duration(frame_count, fps)
    assert result == expected_duration


def test_calculate_duration_with_negative_frame_count_should_raise_validation_error():
    with pytest.raises(ValidationError):
        calculate_duration(-120, 30.0)


def test_calculate_duration_with_frame_count_zero_should_raise_validation_error():
    with pytest.raises(ValidationError):
        calculate_duration(0, 30.0)


def test_calculate_duration_with_frame_count_point_five_should_raise_validation_error():
    with pytest.raises(ValidationError):
        calculate_duration(0.5, 30.0)


def test_calculate_duration_with_negative_fps_should_raise_validation_error():
    with pytest.raises(ValidationError):
        calculate_duration(120, -30.0)


def test_calculate_duration_with_fps_zero_should_raise_validation_error():
    with pytest.raises(ValidationError):
        calculate_duration(120, 0.0)


def test_timedelta_to_timestamp_with_zero_seconds():
    duration = timedelta(seconds=0)
    result = timedelta_to_timestamp(duration)
    assert result == "00:00:00"


def test_timedelta_to_timestamp_with_59_seconds():
    duration = timedelta(seconds=59)
    result = timedelta_to_timestamp(duration)
    assert result == "00:00:59"


def test_timedelta_to_timestamp_with_61_seconds():
    duration = timedelta(seconds=61)
    result = timedelta_to_timestamp(duration)
    assert result == "00:01:01"


def test_timedelta_to_timestamp_with_microseconds_should_truncate():
    duration = timedelta(seconds=61.123456)
    result = timedelta_to_timestamp(duration)
    assert result == "00:01:01"


def test_timedelta_to_timestamp_with_negative_duration_should_raise_value_error():
    with pytest.raises(ValueError) as exc_info:
        duration = timedelta(seconds=-1)
        timedelta_to_timestamp(duration)
    assert "Invalid duration: timedelta must be non-negative" in str(exc_info.value)


def test_convert_bytes_with_bytes():
    bytes_ = 1024
    result = convert_bytes(bytes_)
    assert result == "1.00 KB"


def test_convert_bytes_with_kilobytes():
    kilobytes = 1024**2
    result = convert_bytes(kilobytes)
    assert result == "1.00 MB"


def test_convert_bytes_with_megabytes():
    megabytes = 1024**3
    result = convert_bytes(megabytes)
    assert result == "1.00 GB"


def test_convert_bytes_with_gigabytes():
    gigabytes = 1024**4
    result = convert_bytes(gigabytes)
    assert result == "1.00 TB"


def test_convert_bytes_with_terabytes():
    terabytes = 1024**5
    result = convert_bytes(terabytes)
    assert result == "1.00 PB"


def test_convert_bytes_with_negative_bytes_should_raise_value_error():
    with pytest.raises(ValidationError) as exc_info:
        bytes_ = -1024
        convert_bytes(bytes_)
    assert "Expected positive integer" in str(exc_info.value)


def test_custom_json_encoder_encode_path():
    path = Path("/path/to/file")
    encoded_path = json.dumps(path, cls=CustomJSONEncoder)
    assert encoded_path == '"C:/path/to/file"'


def test_custom_json_encoder_encode_timedelta():
    duration = timedelta(hours=1, minutes=11)
    encoded_duration = json.dumps(duration, cls=CustomJSONEncoder)
    assert encoded_duration == '"1:11:00"'


def test_custom_json_encoder_encode_custom_object():
    custom_obj = ExtractionMethod.AUDIO
    encoded_custom_obj = json.dumps(custom_obj, cls=CustomJSONEncoder)
    assert encoded_custom_obj == '"audio"'


def test_to_json_mixin_to_json_skip_private_keys_is_false():
    @dataclass
    class TestClass(ToJsonMixin):
        a: str
        b: int
        _private: bool = True

    test_obj = TestClass("c", 4)
    expected_json = '{\n  "a": "c",\n  "b": 4,\n  "_private": true\n}'
    assert test_obj.json(skip_private_keys=False) == expected_json


def test_to_json_mixin_to_json_skip_private_keys_is_true():
    @dataclass
    class TestClass(ToJsonMixin):
        a: str
        b: int
        _private: bool = True

    test_obj = TestClass("c", 4)
    expected_json = '{\n  "a": "c",\n  "b": 4\n}'
    assert test_obj.json(skip_private_keys=True) == expected_json
